from pathlib import Path

import click
import yaml

from bagels.config import Config
from bagels.locations import config_file, database_file, set_custom_root


def create_config_file() -> None:
    f = config_file()
    if f.exists():
        return

    try:
        f.touch()
        with open(f, "w") as f:
            yaml.dump(Config.get_default().model_dump(), f)
    except OSError:
        pass


@click.group(invoke_without_command=True)
@click.option(
    "--at",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, path_type=Path),
    help="Specify the path.",
)
@click.pass_context
def cli(ctx, at: click.Path | None):
    """Bagels CLI."""
    if at:
        set_custom_root(at)
    if ctx.invoked_subcommand is None:

        from bagels.models.database.app import init_db

        init_db()

        from bagels.app import App

        app = App()
        app.run()


@cli.command()
@click.argument("thing_to_locate", type=click.Choice(["config", "database"]))
def locate(thing_to_locate: str) -> None:
    if thing_to_locate == "config":
        print("Config file:")
        print(config_file())
    elif thing_to_locate == "database":
        print("Database file:")
        print(database_file())


if __name__ == "__main__":
    cli()



from datetime import datetime
from typing import Tuple, Dict, Any

from textual.widget import Widget

from bagels.forms.form import Form, FormField


def _validate_number(
    value: str, field: FormField, is_float: bool = False
) -> Tuple[bool, str | None]:
    """Validate a number field and return (is_valid, error_message)"""
    if not value:
        if field.is_required:
            return False, "Required"
        return True, None

    # Check if valid number
    if is_float:
        # Allow negative sign at start
        test_value = value.lstrip("-").replace(".", "", 1)
        is_valid = test_value.isdigit()
        type_name = "number"
    else:
        # Allow negative sign at start
        test_value = value.lstrip("-")
        is_valid = test_value.isdigit()
        type_name = "integer"

    if not is_valid:
        return False, f"Must be a {type_name}"

    # Convert to number for comparisons
    num_val = float(value) if is_float else int(value)

    # Check minimum
    if field.min is not None:
        min_val = float(field.min) if is_float else int(field.min)
        if num_val <= min_val:
            return False, f"Must be greater than {field.min}"

    # Check maximum
    if field.max is not None:
        max_val = float(field.max) if is_float else int(field.max)
        if num_val > max_val:
            return False, f"Must be less than {field.max}"

    return True, None


def _validate_date(
    value: str, field: FormField, auto_day: bool = False
) -> Tuple[datetime | None, str | None]:
    """Validate a date field and return (parsed_date, error_message)"""
    if not value or value == "":
        if field.is_required:
            return None, "Required"
        return None, None

    try:
        if auto_day and value.isdigit():
            # Use current month/year if not provided
            this_month = datetime.now().strftime("%m")
            this_year = datetime.now().strftime("%y")
            date = datetime.strptime(f"{value} {this_month} {this_year}", "%d %m %y")
            return date, None
        date = datetime.strptime(value, "%d %m %y")
        return date, None
    except ValueError:
        format_str = "dd (mm) (yy) format." if auto_day else "dd mm yy format"
        return None, f"Must be in {format_str}"


def _validate_autocomplete(
    value: str, held_value: str, field: FormField
) -> Tuple[bool, str | None]:
    """Validate an autocomplete field and return (is_valid, error_message)"""
    if not value and not held_value:
        if field.is_required:
            return False, "Must be selected"
        return True, None

    if not field.options or not field.options.items:
        return True, None

    if field.options.items[0].text:
        # Find matching option
        field_input_value = None
        for item in field.options.items:
            if item.text == value:
                field_input_value = str(item.value)
                break

        # Verify selected value matches entered text
        if field_input_value != str(held_value):
            print(field_input_value, held_value)
            return False, "Invalid selection"

    return True, None


def validateForm(
    formComponent: Widget, formData: Form
) -> Tuple[Dict[str, Any], Dict[str, str], bool]:
    result = {}
    errors = {}
    isValid = True

    for field in formData.fields:
        fieldKey = field.key
        fieldWidget = formComponent.query_one(f"#field-{fieldKey}")
        fieldValue = (
            fieldWidget.heldValue
            if hasattr(fieldWidget, "heldValue")
            else fieldWidget.value
        )

        error = None

        match field.type:
            case "integer":
                is_valid, error = _validate_number(fieldValue, field)
                if is_valid and fieldValue:
                    result[fieldKey] = int(fieldValue)

            case "number":
                is_valid, error = _validate_number(fieldValue, field, is_float=True)
                if is_valid and fieldValue:
                    result[fieldKey] = float(fieldValue)

            case "date":
                date, error = _validate_date(fieldValue, field)
                if date:
                    result[fieldKey] = date

            case "dateAutoDay":
                date, error = _validate_date(fieldValue, field, auto_day=True)
                if date:
                    result[fieldKey] = date

            case "autocomplete":
                is_valid, error = _validate_autocomplete(
                    fieldWidget.value, fieldValue, field
                )
                if is_valid and fieldValue:
                    result[fieldKey] = fieldValue

            case _:
                if not fieldValue and field.is_required:
                    error = "Required"
                else:
                    result[fieldKey] = fieldValue

        if error:
            errors[fieldKey] = error
            isValid = False

    return result, errors, isValid



from datetime import datetime
from enum import Enum
from sqlalchemy import Column, DateTime, Integer, String, Enum as SQLEnum, ForeignKey
from sqlalchemy.orm import relationship
from .database.db import Base


class Nature(Enum):
    WANT = "Want"
    NEED = "Need"
    MUST = "Must"

    def __str__(self):
        return self.value


class Category(Base):
    __tablename__ = "category"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )
    deletedAt = Column(DateTime, nullable=True)

    id = Column(Integer, primary_key=True, index=True)
    parentCategoryId = Column(Integer, ForeignKey("category.id"), nullable=True)
    name = Column(String, nullable=False)
    nature = Column(SQLEnum(Nature), nullable=False)
    color = Column(String, nullable=False)

    records = relationship("Record", back_populates="category")
    parentCategory = relationship(
        "Category", back_populates="subCategories", remote_side=[id]
    )
    subCategories = relationship(
        "Category", back_populates="parentCategory", remote_side=[parentCategoryId]
    )



from datetime import datetime
from sqlalchemy import Column, DateTime, Integer, String, Float, Boolean
from sqlalchemy.orm import relationship
from .database.db import Base


class Account(Base):
    __tablename__ = "account"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )
    deletedAt = Column(DateTime, nullable=True)

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, nullable=False)
    description = Column(String)
    beginningBalance = Column(Float, nullable=False)
    repaymentDate = Column(Integer)

    hidden = Column(Boolean, nullable=False, default=False)

    records = relationship(
        "Record", back_populates="account", foreign_keys="[Record.accountId]"
    )
    transferFromRecords = relationship(
        "Record",
        back_populates="transferToAccount",
        foreign_keys="[Record.transferToAccountId]",
    )
    splits = relationship("Split", back_populates="account")



import copy
from bagels.managers.persons import get_person_by_id
from bagels.forms.form import Form, FormField, Option, Options


class PersonForm:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # ------------ Blueprints ------------ #

    FORM = Form(
        fields=[
            FormField(
                placeholder="Steve",
                title="Name",
                key="name",
                type="string",
                is_required=True,
            ),
        ]
    )
    # ------------- Builders ------------- #

    def get_filled_form(self, personId: int):
        form = copy.deepcopy(self.FORM)
        person = get_person_by_id(personId)
        for field in form.fields:
            value = getattr(person, field.key)
            field.default_value = str(value) if value is not None else ""
        return form

    def get_form(self):
        return copy.deepcopy(self.FORM)



import pytest
from sqlalchemy import create_engine, delete
from sqlalchemy.orm import sessionmaker
from bagels.models.database.db import Base
from bagels.models.record_template import RecordTemplate
from bagels.models.account import Account
from bagels.models.category import Category, Nature
from bagels.managers import record_templates

# Setup test database
@pytest.fixture(scope="function")
def engine():
    engine = create_engine("sqlite:///:memory:")
    Base.metadata.create_all(engine)
    yield engine
    Base.metadata.drop_all(engine)

@pytest.fixture(scope="function")
def session(engine):
    Session = sessionmaker(bind=engine)
    session = Session()
    yield session
    session.close()

@pytest.fixture(autouse=True)
def setup_test_engine(monkeypatch, engine):
    # Replace the global engine with our test engine
    import bagels.managers.record_templates as rt
    rt.Session = sessionmaker(bind=engine)
    yield

@pytest.fixture
def account(session):
    account = Account(name="Test Account", beginningBalance=1000.0)
    session.add(account)
    session.commit()
    return account

@pytest.fixture
def category(session):
    category = Category(
        name="Test Category",
        nature=Nature.WANT,
        color="#000000"
    )
    session.add(category)
    session.commit()
    return category

@pytest.fixture
def template_data(account, category):
    return {
        "label": "Test Template",
        "amount": 100.0,
        "accountId": account.id,
        "categoryId": category.id,
        "isIncome": False,
        "isTransfer": False,
        "transferToAccountId": None,
    }

def test_create_template(template_data):
    template = record_templates.create_template(template_data)
    assert template.label == template_data["label"]
    assert template.amount == template_data["amount"]
    assert template.accountId == template_data["accountId"]
    assert template.categoryId == template_data["categoryId"]
    assert template.order == 1  # First template should have order 1

def test_create_multiple_templates(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)
    assert template1.order == 1
    assert template2.order == 2

def test_get_all_templates(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)
    templates = record_templates.get_all_templates()
    assert len(templates) == 2
    assert templates[0].id == template1.id
    assert templates[1].id == template2.id
    assert templates[0].order < templates[1].order

def test_get_template_by_id(template_data):
    template = record_templates.create_template(template_data)
    retrieved = record_templates.get_template_by_id(template.id)
    assert retrieved is not None
    assert retrieved.id == template.id
    assert retrieved.label == template.label

def test_get_nonexistent_template():
    template = record_templates.get_template_by_id(999)
    assert template is None

def test_get_adjacent_template(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)
    template3 = record_templates.create_template(template_data)

    # Test getting next template
    next_id = record_templates.get_adjacent_template(template1.id, "next")
    assert next_id == template2.id

    # Test getting previous template
    prev_id = record_templates.get_adjacent_template(template2.id, "previous")
    assert prev_id == template1.id

    # Test getting next template for last template
    next_id = record_templates.get_adjacent_template(template3.id, "next")
    assert next_id == -1

    # Test getting previous template for first template
    prev_id = record_templates.get_adjacent_template(template1.id, "previous")
    assert prev_id == -1

def test_update_template(template_data):
    template = record_templates.create_template(template_data)
    updates = {
        "label": "Updated Template",
        "amount": 200.0
    }
    updated = record_templates.update_template(template.id, updates)
    assert updated.label == updates["label"]
    assert updated.amount == updates["amount"]
    assert updated.accountId == template.accountId  # Unchanged fields remain the same

def test_update_nonexistent_template(template_data):
    updated = record_templates.update_template(999, {"label": "New Label"})
    assert updated is None

def test_swap_template_order(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)
    template3 = record_templates.create_template(template_data)

    # Test swapping with next template
    swapped = record_templates.swap_template_order(template1.id, "next")
    templates = record_templates.get_all_templates()
    assert templates[0].id == template2.id
    assert templates[1].id == template1.id

    # Test swapping with previous template
    swapped = record_templates.swap_template_order(template1.id, "previous")
    templates = record_templates.get_all_templates()
    assert templates[0].id == template1.id
    assert templates[1].id == template2.id

def test_swap_template_order_edge_cases(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)

    # Test swapping first template with previous (should fail gracefully)
    swapped = record_templates.swap_template_order(template1.id, "previous")
    templates = record_templates.get_all_templates()
    assert templates[0].id == template1.id

    # Test swapping last template with next (should fail gracefully)
    swapped = record_templates.swap_template_order(template2.id, "next")
    templates = record_templates.get_all_templates()
    assert templates[1].id == template2.id

def test_delete_template(template_data):
    template1 = record_templates.create_template(template_data)
    template2 = record_templates.create_template(template_data)
    template3 = record_templates.create_template(template_data)

    # Delete middle template
    success = record_templates.delete_template(template2.id)
    assert success is True

    # Check remaining templates
    templates = record_templates.get_all_templates()
    assert len(templates) == 2
    assert templates[0].id == template1.id
    assert templates[1].id == template3.id
    assert templates[0].order == 1  # Orders should be resequenced
    assert templates[1].order == 2

def test_delete_nonexistent_template():
    success = record_templates.delete_template(999)
    assert success is False

def test_template_validation(template_data, account):
    # Test negative amount
    invalid_data = template_data.copy()
    invalid_data["amount"] = -100.0
    with pytest.raises(Exception):  # SQLAlchemy will raise an exception due to CheckConstraint
        record_templates.create_template(invalid_data)

    # Test transfer template
    transfer_data = template_data.copy()
    transfer_data["isTransfer"] = True
    transfer_data["transferToAccountId"] = account.id
    template = record_templates.create_template(transfer_data)
    assert template.isTransfer is True
    assert template.transferToAccountId == account.id

    # Test invalid transfer (transfer with income)
    invalid_transfer = template_data.copy()
    invalid_transfer["isTransfer"] = True
    invalid_transfer["isIncome"] = True
    invalid_transfer["transferToAccountId"] = account.id
    with pytest.raises(Exception):  # SQLAlchemy will raise an exception due to CheckConstraint
        record_templates.create_template(invalid_transfer)



from datetime import datetime
from sqlalchemy import Column, DateTime, Integer, Float, Boolean, ForeignKey
from sqlalchemy.orm import relationship
from .database.db import Base


class Split(Base):
    __tablename__ = "split"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )

    id = Column(Integer, primary_key=True, index=True)
    recordId = Column(
        Integer, ForeignKey("record.id", ondelete="CASCADE"), nullable=False
    )
    amount = Column(Float, nullable=False)
    personId = Column(Integer, ForeignKey("person.id"), nullable=False)
    isPaid = Column(Boolean, nullable=False, default=False)
    paidDate = Column(DateTime, nullable=True)
    accountId = Column(Integer, ForeignKey("account.id"), nullable=True)

    record = relationship("Record", foreign_keys=[recordId], back_populates="splits")
    person = relationship("Person", foreign_keys=[personId], back_populates="splits")
    account = relationship("Account", foreign_keys=[accountId], back_populates="splits")



from sqlalchemy.orm import sessionmaker
from bagels.models.split import Split
from bagels.models.database.app import db_engine

Session = sessionmaker(bind=db_engine)


def create_split(data):
    session = Session()
    try:
        new_split = Split(**data)
        session.add(new_split)
        session.commit()
        session.refresh(new_split)
        session.expunge(new_split)
        return new_split
    finally:
        session.close()


def get_splits_by_record_id(record_id):
    session = Session()
    try:
        return session.query(Split).filter_by(recordId=record_id).all()
    finally:
        session.close()


def get_split_by_id(split_id):
    session = Session()
    try:
        return session.query(Split).get(split_id)
    finally:
        session.close()


def update_split(split_id, updated_data):
    session = Session()
    try:
        split = session.query(Split).get(split_id)
        if split:
            for key, value in updated_data.items():
                setattr(split, key, value)
            session.commit()
        return split
    finally:
        session.close()


def delete_split(split_id):
    session = Session()
    try:
        split = session.query(Split).get(split_id)
        if split:
            session.delete(split)
            session.commit()
        return split
    finally:
        session.close()


def delete_splits_by_record_id(record_id):
    session = Session()
    try:
        session.query(Split).filter_by(recordId=record_id).delete()
        session.commit()
    finally:
        session.close()



from pydantic import BaseModel
from rich.color import Color as RichColor
from textual.app import ComposeResult
from textual.color import Color
from textual.containers import Container
from textual.widgets import Label, Static


class PercentageBarItem(BaseModel):
    name: str
    count: int
    color: str


class PercentageBar(Static):
    DEFAULT_CSS = """
    PercentageBar {
        layout: vertical;
        width: 1fr;
        height: auto;
    }
    
    PercentageBar > .bar-container {
        layout: horizontal;
        height: 1;
        width: 1fr;
        
        .bar {
            layout: horizontal;
            height: 1;
            width: 1fr;
            
            .empty-bar {
                hatch: right $panel-lighten-2;
                
                Label {
                    dock: left;
                    align-horizontal: center;
                    padding: 0 1 0 1;
                }
            }
        }
    }
    
    
    PercentageBar > .labels-container {
        layout: vertical;
        height: auto;
        margin-top: 1;
        
        .bar-label {
            layout: horizontal;
            height: 1;
        }
        
        .bar-label > .percentage {
            margin-left: 2;
            color: $primary-background-lighten-3;
        }
    }
    
    """

    items: list[PercentageBarItem] = []

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.rounded = False

    def on_mount(self) -> None:
        self.rebuild()

    def set_items(self, items: list[PercentageBarItem]) -> None:
        self.items = items
        self.rebuild()

    #  50%  50% 
    # ^-----======^
    def rebuild(self) -> None:
        # we first remove all existing items and labels
        for item in self.query(".bar-item"):
            item.remove()
        labels = self.query(".bar-label")
        labels_count = len(labels)
        items_count = len(self.items)

        prev_empty_bar = self.bar.query(".empty-bar")
        if len(self.items) == 0:
            if self.rounded:
                self.bar_start.styles.display = "none"
                self.bar_end.styles.display = "none"
            if len(prev_empty_bar) == 0:
                empty_bar = Container(Label("No data to display"), classes="empty-bar")
                self.bar.mount(empty_bar)
        else:
            if self.rounded:
                self.bar_start.styles.display = "block"
                self.bar_end.styles.display = "block"
            if len(prev_empty_bar) > 0:
                prev_empty_bar[0].remove()

        to_remove_count = labels_count - items_count
        if to_remove_count > 0:
            for i in range(to_remove_count):
                labels[i + items_count].remove()
        # we calculate the appropriate width for each item, with last item taking remaining space
        total = sum(item.count for item in self.items)
        for i, item in enumerate(self.items):
            item_widget = Static(" ", classes="bar-item")
            color = item.color
            background_color = Color.from_rich_color(RichColor.parse(item.color)).hex
            # assign start and end colors
            if self.rounded:
                if i == 0:
                    self.bar_start.styles.color = background_color
                if i == len(self.items) - 1:
                    self.bar_end.styles.color = background_color
            # calculate percentage
            percentage = round((item.count / total) * 100)
            if (
                i + 1 > labels_count
            ):  # if we have more items than labels, we create a new label
                label_widget = Container(
                    Label(f"[{color}]●[/{color}] {item.name}", classes="name"),
                    Label(f"{percentage}%", classes="percentage"),
                    classes="bar-label",
                )
                self.labels_container.mount(label_widget)
            else:
                label = labels[i]
                label.query_one(".name").update(f"[{color}]●[/{color}] {item.name}")
                label.query_one(".percentage").update(f"{percentage}%")

            width = f"{percentage}%"
            if i == len(self.items) - 1:
                # Last item takes remaining space
                width = "1fr"

            item_widget.styles.width = width
            item_widget.styles.background = background_color
            if self.rounded:
                if i > 0:
                    prev_background_color = Color.from_rich_color(
                        RichColor.parse(self.items[i - 1].color)
                    ).hex
                    item_widget.update(
                        f"[{prev_background_color} on {background_color}][/{prev_background_color} on {background_color}]"
                    )

            self.bar.mount(item_widget)

    def compose(self) -> ComposeResult:
        self.bar_start = Label("", classes="bar-start")
        self.bar = Container(classes="bar")
        self.bar_end = Label("", classes="bar-end")
        self.labels_container = Container(classes="labels-container")
        with Container(classes="bar-container"):
            if self.rounded:
                yield self.bar_start
            yield self.bar
            if self.rounded:
                yield self.bar_end
        yield self.labels_container



from textual.app import ComposeResult
from textual.widgets import Static


class EmptyIndicator(Static):
    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def compose(self) -> ComposeResult:
        yield Static(self.message, id="empty-indicator")



import yaml
from pathlib import Path
from sqlalchemy import create_engine, inspect, text
from sqlalchemy.orm import sessionmaker

from bagels.models.category import Nature
from bagels.locations import database_file

# -------- create all imports -------- #
from bagels.models.account import Account
from bagels.models.category import Category
from bagels.models.person import Person
from bagels.models.record import Record
from bagels.models.record_template import RecordTemplate
from bagels.models.split import Split

from bagels.models.database.db import Base

db_engine = create_engine(f"sqlite:///{database_file().resolve()}")
Session = sessionmaker(bind=db_engine)


def _create_outside_source_account(session):
    outside_account = session.query(Account).filter_by(name="Outside source").first()
    if not outside_account:
        outside_account = Account(
            name="Outside source",
            description="Default account for external transactions",
            beginningBalance=0.0,
            hidden=True,
        )
        session.add(outside_account)
        session.commit()


def _create_default_categories(session):
    category_count = session.query(Category).count()
    if category_count > 0:
        return

    yaml_path = (
        Path(__file__).parent.parent.parent / "static" / "default_categories.yaml"
    )

    with open(yaml_path, "r") as file:
        default_categories = yaml.safe_load(file)

    for category in default_categories:
        parent = Category(
            name=category["name"],
            nature=getattr(Nature, category["nature"]),
            color=category["color"],
            parentCategoryId=None,
        )
        session.add(parent)
        session.commit()

        for subcategory in category["subcategories"]:
            child = Category(
                name=subcategory["name"],
                nature=getattr(Nature, subcategory["nature"]),
                color=category["color"],
                parentCategoryId=parent.id,
            )
            session.add(child)
            session.commit()


def _sync_database_schema():
    try:
        inspector = inspect(db_engine)
        existing_tables = inspector.get_table_names()

        for table in Base.metadata.tables.values():
            if table.name not in existing_tables:
                table.create(db_engine)
            else:
                existing_columns = {
                    col["name"] for col in inspector.get_columns(table.name)
                }
                model_columns = {col.name for col in table.columns}

                for column_name in model_columns - existing_columns:
                    column = table.columns[column_name]
                    with db_engine.begin() as conn:
                        conn.execute(
                            text(
                                f'ALTER TABLE {table.name} ADD COLUMN "{column_name}" '
                                f"{column.type} "
                                f"{'NOT NULL' if not column.nullable else ''} "
                                f"{'DEFAULT ' + str(column.default.arg) if column.default is not None else ''}"
                            )
                        )
    except Exception as e:
        raise Exception(f"Failed to sync database schema: {str(e)}")


def init_db():
    _sync_database_schema()
    Base.metadata.create_all(db_engine)
    session = Session()
    _create_outside_source_account(session)
    _create_default_categories(session)
    session.close()


def wipe_database():
    Base.metadata.drop_all(db_engine)
    _sync_database_schema()
    Base.metadata.create_all(db_engine)
    session = Session()
    _create_outside_source_account(session)
    _create_default_categories(session)
    session.close()



from datetime import datetime, timedelta

from bagels.config import CONFIG


def format_date_to_readable(date) -> str:
    today = datetime.now().date()
    date = date.date() if isinstance(date, datetime) else date

    if date == today:
        return "Today"
    elif date == today - timedelta(days=1):
        return "Yesterday"

    # Get start of current week (Monday)
    start_of_week = today - timedelta(days=today.weekday())
    # Get end of current week (Sunday)
    end_of_week = start_of_week + timedelta(days=6)

    if start_of_week <= date <= end_of_week:
        return date.strftime("%A")
    else:
        return date.strftime(CONFIG.defaults.date_format)


def format_period_to_readable(filter: dict) -> str:
    offset = filter["offset"]
    offset_type = filter["offset_type"]
    first_day_of_week = CONFIG.defaults.first_day_of_week
    if offset_type == "day":
        return format_date_to_readable(datetime.now() + timedelta(days=offset))
    match offset:
        case 0:
            return f"This {offset_type.title()}"
        case -1:
            return f"Last {offset_type.title()}"
        case _:
            now = datetime.now()
            match offset_type:
                case "year":
                    target_year = now.year + offset
                    return f"{target_year}"
                case "month":
                    target_month = now.month + offset
                    target_year = now.year + (target_month - 1) // 12
                    target_month = ((target_month - 1) % 12) + 1
                    return f"{datetime(target_year, target_month, 1).strftime('%B %Y')}"
                case "week":
                    target_date = now + timedelta(weeks=offset)
                    # Calculate days to first day of week based on first_day_of_week setting
                    days_to_first = (target_date.weekday() - first_day_of_week) % 7
                    start = target_date - timedelta(days=days_to_first)
                    end = start + timedelta(days=6)
                    return f"{start.strftime('%d %b')} - {end.strftime('%d %b')}"
                case "day":
                    target_date = now + timedelta(days=offset)
                    return f"{target_date.strftime('%d %B %Y')}"



from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Static

from bagels.components.datatable import DataTable
from bagels.components.indicators import EmptyIndicator
from bagels.modals.base_widget import ModalContainer
from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.input import InputModal

from bagels.config import CONFIG
from bagels.managers.categories import (
    create_category,
    delete_category,
    get_all_categories_tree,
    get_categories_count,
    get_category_by_id,
    update_category,
)
from bagels.forms.category_form import CategoryForm


class CategoriesModal(ModalScreen[str | Widget | None]):

    COLUMNS = ("", "Name", "Nature")

    BINDINGS = [
        Binding(
            CONFIG.hotkeys.categories.browse_defaults,
            "browse_defaults",
            "Browse",
            False,
        ),
        Binding(CONFIG.hotkeys.new, "new_category", "Add"),
        Binding(
            CONFIG.hotkeys.categories.new_subcategory,
            "new_subcategory",
            "Add Subcategory",
        ),
        Binding(CONFIG.hotkeys.edit, "edit_category", "Edit"),
        Binding(CONFIG.hotkeys.delete, "delete_category", "Delete"),
    ]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="categories-modal-screen", classes="modal-screen"
        )
        self.title = "Manage your categories"

    # --------------- Hooks -------------- #

    def on_mount(self) -> None:
        categories_is_empty = get_categories_count() == 0
        self.rebuild()

    def on_key(self, event: events.Key) -> None:
        if event.key == "escape":
            self.dismiss()

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        if event.row_key:
            self.current_row = event.row_key.value

    # region Builders
    # ------------- Builders ------------- #

    def rebuild(self) -> None:
        table: DataTable = self.query_one("#categories-table")
        empty_indicator: Static = self.query_one("#empty-indicator")

        table.clear()
        if not table.columns:
            table.add_columns(*self.COLUMNS)

        categories = get_all_categories_tree()
        if categories:
            for category, node in categories:
                table.add_row(
                    node, category.name, category.nature.value, key=category.id
                )
            table.zebra_stripes = True
            table.focus()
        else:
            self.current_row = None

        empty_indicator.display = not categories

    # region Helpers
    # -------------- Helpers ------------- #

    def _notify_no_categories(self) -> None:
        self.app.notify(
            title="Error",
            message="Category must be selected for this action.",
            severity="error",
            timeout=2,
        )

    # def new_binding(self, binding: Binding) -> None:
    #     self._bindings.key_to_bindings.setdefault(binding.key, []).append(binding)

    # def remove_binding(self, key: str) -> None:
    #     self._bindings.key_to_bindings.pop(key, None)

    # region callbacks
    # ------------- Callbacks ------------ #

    def action_new_category(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_category(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Category created",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        self.app.push_screen(
            InputModal("New Category", CategoryForm().get_form()), callback=check_result
        )

    def action_new_subcategory(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_result(result) -> None:
            if result:
                try:
                    create_category(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Subcategory created",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        parent_category_id = self.current_row
        subcategory_form = CategoryForm().get_subcategory_form(parent_category_id)
        parent_category = get_category_by_id(parent_category_id)
        self.app.push_screen(
            InputModal(f"New Subcategory of {parent_category.name}", subcategory_form),
            callback=check_result,
        )

    def action_delete_category(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_delete(result) -> None:
            if result:
                try:
                    delete_category(self.current_row)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                self.rebuild()

        self.app.push_screen(
            ConfirmationModal("Are you sure you want to delete this record?"),
            check_delete,
        )

    def action_edit_category(self) -> None:
        if not self.current_row:
            self._notify_no_categories()
            return

        def check_result(result) -> None:
            if result:
                try:
                    update_category(self.current_row, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Category {result['name']} updated",
                        severity="information",
                        timeout=3,
                    )
                    self.rebuild()

        filled_form = CategoryForm().get_filled_form(self.current_row)
        self.app.push_screen(
            InputModal("Edit Category", filled_form), callback=check_result
        )

    # region View
    # --------------- View --------------- #
    def compose(self) -> ComposeResult:
        yield ModalContainer(
            DataTable(
                id="categories-table",
                cursor_type="row",
                cursor_foreground_priority=True,
            ),
            EmptyIndicator("No categories"),
        )



from datetime import datetime
from sqlalchemy import (
    Column,
    DateTime,
    Integer,
    String,
    Float,
    Boolean,
    ForeignKey,
    CheckConstraint,
)
from sqlalchemy.orm import relationship
from .database.db import Base


class Record(Base):
    __tablename__ = "record"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )

    id = Column(Integer, primary_key=True, index=True)
    label = Column(String, nullable=False)
    amount = Column(Float, CheckConstraint("amount > 0"), nullable=False)
    date = Column(DateTime, nullable=False, default=datetime.now)
    accountId = Column(Integer, ForeignKey("account.id"), nullable=False)
    categoryId = Column(Integer, ForeignKey("category.id"), nullable=True)

    tags = Column(String, nullable=True)  # unimplemented
    isInProgress = Column(Boolean, nullable=False, default=False)  # unimplemented

    # if record adds money to account
    isIncome = Column(Boolean, nullable=False, default=False)
    # if record is transfer to this account
    isTransfer = Column(
        Boolean,
        CheckConstraint("(isTransfer = FALSE) OR (isIncome = FALSE)"),
        nullable=False,
        default=False,
    )
    transferToAccountId = Column(Integer, ForeignKey("account.id"), nullable=True)

    account = relationship(
        "Account", foreign_keys=[accountId], back_populates="records"
    )
    category = relationship("Category", back_populates="records")
    transferToAccount = relationship(
        "Account",
        foreign_keys=[transferToAccountId],
        back_populates="transferFromRecords",
    )
    splits = relationship(
        "Split", back_populates="record", cascade="all, delete-orphan"
    )



import pytest
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from bagels.models.database.db import Base
from bagels.models.account import Account
from bagels.managers import accounts

@pytest.fixture(scope="function")
def test_db():
    # Create in-memory SQLite database
    engine = create_engine("sqlite:///:memory:")
    
    # Create all tables
    Base.metadata.create_all(engine)
    
    # Create a new session factory bound to the engine
    accounts.Session = sessionmaker(bind=engine)
    
    yield engine
    
    # Clean up
    Base.metadata.drop_all(engine)

def test_create_account(test_db):
    # Test data
    account_data = {
        "name": "Test Account",
        "description": "Test Description",
        "beginningBalance": 1000.0,
        "repaymentDate": 15,
        "hidden": False
    }
    
    # Create account
    new_account = accounts.create_account(account_data)
    
    # Assertions
    assert new_account is not None
    assert new_account.name == "Test Account"
    assert new_account.description == "Test Description"
    assert new_account.beginningBalance == 1000.0
    assert new_account.repaymentDate == 15
    assert new_account.hidden is False
    assert new_account.id is not None
    assert new_account.createdAt is not None
    assert new_account.updatedAt is not None
    assert new_account.deletedAt is None

def test_get_all_accounts(test_db):
    # Create test data
    account_data1 = {
        "name": "Account 1",
        "beginningBalance": 1000.0,
        "hidden": False
    }
    account_data2 = {
        "name": "Account 2",
        "beginningBalance": 2000.0,
        "hidden": True
    }
    
    accounts.create_account(account_data1)
    accounts.create_account(account_data2)
    
    # Get all non-hidden accounts
    visible_accounts = accounts.get_all_accounts(get_hidden=False)
    assert len(visible_accounts) == 1
    assert visible_accounts[0].name == "Account 1"
    
    # Get all accounts including hidden
    all_accounts = accounts.get_all_accounts(get_hidden=True)
    assert len(all_accounts) == 2
    assert {a.name for a in all_accounts} == {"Account 1", "Account 2"}

def test_get_accounts_count(test_db):
    # Create test data
    account_data1 = {
        "name": "Account 1",
        "beginningBalance": 1000.0,
        "hidden": False
    }
    account_data2 = {
        "name": "Account 2",
        "beginningBalance": 2000.0,
        "hidden": True
    }
    
    accounts.create_account(account_data1)
    accounts.create_account(account_data2)
    
    # Test counts
    assert accounts.get_accounts_count(get_hidden=False) == 1
    assert accounts.get_accounts_count(get_hidden=True) == 2

def test_get_account_by_id(test_db):
    # Create test data
    account_data = {
        "name": "Test Account",
        "beginningBalance": 1000.0,
        "hidden": False
    }
    new_account = accounts.create_account(account_data)
    
    # Get account by ID
    retrieved_account = accounts.get_account_by_id(new_account.id)
    
    # Assertions
    assert retrieved_account is not None
    assert retrieved_account.id == new_account.id
    assert retrieved_account.name == "Test Account"
    assert retrieved_account.beginningBalance == 1000.0

def test_get_nonexistent_account(test_db):
    # Try to get an account with non-existent ID
    retrieved_account = accounts.get_account_by_id(999)
    
    # Assertions
    assert retrieved_account is None

def test_update_account(test_db):
    # Create test data
    account_data = {
        "name": "Test Account",
        "beginningBalance": 1000.0,
        "hidden": False
    }
    new_account = accounts.create_account(account_data)
    
    # Update data
    update_data = {
        "name": "Updated Account",
        "description": "New Description",
        "hidden": True
    }
    updated_account = accounts.update_account(new_account.id, update_data)
    
    # Assertions
    assert updated_account is not None
    assert updated_account.name == "Updated Account"
    assert updated_account.description == "New Description"
    assert updated_account.hidden is True
    assert updated_account.beginningBalance == 1000.0  # Unchanged
    assert updated_account.id == new_account.id

def test_update_nonexistent_account(test_db):
    # Try to update non-existent account
    update_data = {"name": "Updated Account"}
    updated_account = accounts.update_account(999, update_data)
    
    # Assertions
    assert updated_account is None

def test_delete_account(test_db):
    # Create test data
    account_data = {
        "name": "Test Account",
        "beginningBalance": 1000.0,
        "hidden": False
    }
    new_account = accounts.create_account(account_data)
    
    # Delete account
    result = accounts.delete_account(new_account.id)
    
    # Assertions
    assert result is True
    
    # Check that the account is marked as deleted
    deleted_account = accounts.get_account_by_id(new_account.id)
    assert deleted_account is not None
    assert deleted_account.deletedAt is not None
    
    # Check that it's not returned in get_all_accounts
    all_accounts = accounts.get_all_accounts(get_hidden=True)
    assert len(all_accounts) == 0

def test_delete_nonexistent_account(test_db):
    # Try to delete non-existent account
    result = accounts.delete_account(999)
    
    # Assertions
    assert result is False



from functools import partial
from typing import TYPE_CHECKING, cast

from textual.command import DiscoveryHit, Hit, Hits, Provider
from textual.types import IgnoreReturnCallbackType

from bagels.models.database.app import wipe_database
from bagels.managers.samples import create_sample_entries

if TYPE_CHECKING:
    from bagels.app import App


class AppProvider(Provider):
    @property
    def commands(
        self,
    ) -> tuple[tuple[str, IgnoreReturnCallbackType, str, bool], ...]:
        app = self.app

        commands_to_show: list[tuple[str, IgnoreReturnCallbackType, str, bool]] = [
            ("app: quit", app.action_quit, "Quit App", True),
            (
                "dev: create sample entries",
                self._action_create_sample_entries,
                "Create sample entries defined in static/sample_entries.yaml",
                False,
            ),
            (
                "dev: wipe database",
                self._action_wipe_database,
                "Delete everything from the database",
                False,
            ),
            *self.get_theme_commands(),
        ]

        return tuple(commands_to_show)

    async def discover(self) -> Hits:
        """Handle a request for the discovery commands for this provider.

        Yields:
            Commands that can be discovered.
        """
        for name, runnable, help_text, show_discovery in self.commands:
            if show_discovery:
                yield DiscoveryHit(
                    name,
                    runnable,
                    help=help_text,
                )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The user input to be matched.

        Yields:
            Command hits for use in the command palette.
        """
        matcher = self.matcher(query)
        for name, runnable, help_text, _ in self.commands:
            if (match := matcher.match(name)) > 0:
                yield Hit(
                    match,
                    matcher.highlight(name),
                    runnable,
                    help=help_text,
                )

    def get_theme_commands(
        self,
    ) -> tuple[tuple[str, IgnoreReturnCallbackType, str, bool], ...]:
        app = self.app
        return tuple(self.get_theme_command(theme) for theme in app.themes)

    def get_theme_command(
        self, theme_name: str
    ) -> tuple[str, IgnoreReturnCallbackType, str, bool]:
        return (
            f"theme: {theme_name}",
            partial(self.app.command_theme, theme_name),
            f"Set the theme to {theme_name}",
            True,
        )

    @property
    def app(self) -> "App":
        return cast("App", self.screen.app)

    def _action_create_sample_entries(self) -> None:
        create_sample_entries()
        self.app.refresh(layout=True, recompose=True)

    def _action_wipe_database(self) -> None:
        wipe_database()
        self.app.refresh(layout=True, recompose=True)
        # def check_delete(result) -> None:

        # self.app.push_screen(ConfirmationModal(
        #     message="Are you sure you want to wipe the database?",
        # ), callback=check_delete)



from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.widgets import (
    Input,
    Label,
)

from bagels.components.autocomplete import AutoComplete, Dropdown, DropdownItem
from bagels.components.fields import Fields
from bagels.config import CONFIG
from bagels.modals.input import InputModal
from bagels.managers.accounts import get_all_accounts_with_balance
from bagels.managers.persons import create_person, get_all_persons
from bagels.forms.form import Form, Option
from bagels.forms.record_forms import RecordForm
from bagels.utils.validation import validateForm
from bagels.modals.base_widget import ModalContainer


class RecordModal(InputModal):

    isEditing = False

    BINDINGS = [
        Binding(
            CONFIG.hotkeys.record_modal.new_split,
            "add_split",
            "Add split",
            priority=True,
        ),
        Binding(
            CONFIG.hotkeys.record_modal.new_paid_split,
            "add_paid_split",
            "Add paid split",
            priority=True,
        ),
        Binding(
            CONFIG.hotkeys.record_modal.delete_last_split,
            "delete_last_split",
            "Delete last split",
            priority=True,
        ),
    ]

    def __init__(
        self,
        title: str,
        form: Form = Form(),
        splitForm: Form = Form(),
        isEditing: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(title, form, *args, **kwargs)
        self.record_form = RecordForm()
        self.splitForm = splitForm
        self.isEditing = isEditing
        if isEditing:
            self._bindings.key_to_bindings.clear()
            self.refresh_bindings()
        self.splitFormOneLength = len(self.record_form.get_split_form(0, False))
        self.splitCount = int(len(splitForm) / self.splitFormOneLength)
        self.persons = get_all_persons()
        self.accounts = get_all_accounts_with_balance()
        self.total_amount = 0
        self.split_total = Label("", id="split-total")

    def on_mount(self):
        self._update_split_total()
        if self.splitCount > 0:
            self._update_split_total_visibility(True)

    # -------------- Helpers ------------- #

    def _get_splits_from_result(self, resultForm: dict):
        splits = []
        for i in range(0, self.splitCount):
            splits.append(
                {
                    "personId": resultForm[f"personId-{i}"],
                    "amount": resultForm[f"amount-{i}"],
                    "isPaid": resultForm[f"isPaid-{i}"],
                    "accountId": resultForm[f"accountId-{i}"],
                    "paidDate": resultForm[f"paidDate-{i}"],
                }
            )
        return splits

    def _update_split_total(self, update_new: bool = True):
        my_amount = self.query_one("#field-amount").value
        total = float(my_amount) if my_amount else 0
        if update_new:
            for i in range(0, self.splitCount):
                amount = self.query_one(f"#field-amount-{i}").value
                total += float(amount) if amount else 0
        self.total_amount = total
        if self.splitCount > 0:
            self.split_total.update(
                f"Total amount: [bold yellow]{total:.2f}[/bold yellow]"
            )

    def _get_split_widget(self, index: int, fields: Form, isPaid: bool):
        widget = Container(Fields(fields), id=f"split-{index}", classes="split")
        widget.border_title = "> Paid split <" if isPaid else "> Split <"
        return widget

    def _get_init_split_widgets(self):
        widgets = []
        for i in range(0, self.splitCount):
            oneSplitForm = Form(
                fields=self.splitForm.fields[
                    i * self.splitFormOneLength : (i + 1) * self.splitFormOneLength
                ]
            )
            # Find the isPaid field in the form fields for this split
            isPaid = False
            for field in oneSplitForm.fields:
                if field.key == f"isPaid-{i}":
                    isPaid = field.default_value
                    break
            widgets.append(self._get_split_widget(i, oneSplitForm, isPaid))
        return widgets

    def _update_split_total_visibility(self, mount: bool):
        if mount:
            self.query_one(".container").mount(self.split_total)
        else:
            self.split_total.remove()

    def _update_errors(self, errors: dict):
        previousErrors = self.query(".error")
        for error in previousErrors:
            error.remove()
        for key, value in errors.items():
            field = self.query_one(f"#row-field-{key}")
            field.mount(Label(value, classes="error"))

    def on_auto_complete_created(self, event: AutoComplete.Created) -> None:
        name = event.item.create_option_text
        person = create_person({"name": name})
        for field in self.splitForm.fields:
            if field.key.startswith("personId"):
                field.options.items.append(Option(text=person.name, value=person.id))
        # update all person dropdowns with the new person
        for i in range(0, self.splitCount):
            dropdown: Dropdown = self.query_one(f"#dropdown-personId-{i}")
            dropdown.items.append(DropdownItem(person.name, "", ""))
        # set heldValue for the AutoComplete's input
        event.input.heldValue = person.id

    # def on_auto

    # ------------- Callbacks ------------ #

    def on_input_changed(self, event: Input.Changed):
        if event.input.id.startswith("field-amount"):
            self._update_split_total()

    def action_add_paid_split(self):
        self.action_add_split(paid=True)

    def action_add_split(self, paid: bool = False):
        splits_container = self.query_one("#splits-container", Container)
        current_split_index = self.splitCount
        new_split_form = self.record_form.get_split_form(current_split_index, paid)
        for field in new_split_form.fields:
            self.splitForm.fields.append(field)
        splits_container.mount(
            self._get_split_widget(current_split_index, new_split_form, paid)
        )
        # Use call_after_refresh to ensure the mount is complete
        splits_container.call_after_refresh(
            lambda: self.query_one(f"#field-personId-{current_split_index}").focus()
        )
        self.splitCount += 1
        if self.splitCount == 1:
            self._update_split_total_visibility(True)
            self._update_split_total(update_new=False)

    def action_delete_last_split(self):
        if self.splitCount > 0:
            self.query_one(f"#split-{self.splitCount - 1}").remove()
            self.query_one(
                f"#dropdown-personId-{self.splitCount - 1}"
            ).remove()  # idk why this is needed
            for i in range(self.splitFormOneLength):
                self.splitForm.fields.pop()
            self.splitCount -= 1
            if self.splitCount == 0:
                self._update_split_total_visibility(False)

    def action_submit(self):
        # We set the amount field to the total amount for the form to read the value
        input: Input = self.query_one("#field-amount")
        input.__setattr__("heldValue", str(self.total_amount))

        resultRecordForm, errors, isValid = validateForm(self, self.form)
        resultSplitForm, errorsSplit, isValidSplit = validateForm(self, self.splitForm)
        if isValid and isValidSplit:
            resultSplits = self._get_splits_from_result(resultSplitForm)
            self.dismiss({"record": resultRecordForm, "splits": resultSplits})
            return
        self._update_errors({**errors, **errorsSplit})
        # Remove the custom value we set for the field if not valid
        input.__setattr__("heldValue", None)

    # -------------- Compose ------------- #

    def compose(self) -> ComposeResult:
        yield ModalContainer(
            Fields(self.form),
            Container(*self._get_init_split_widgets(), id="splits-container"),
        )



from datetime import datetime, timedelta

from rich.text import Text
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.reactive import reactive
from textual.widgets import Static

from textual.widgets import Button
from bagels.components.datatable import DataTable
from bagels.components.indicators import EmptyIndicator
from bagels.modals.confirmation import ConfirmationModal
from bagels.modals.record import RecordModal
from bagels.modals.transfer import TransferModal
from bagels.modals.input import InputModal
from bagels.config import CONFIG
from bagels.managers.persons import (
    get_person_by_id,
    get_persons_with_splits,
    update_person,
)
from bagels.managers.records import (
    create_record,
    create_record_and_splits,
    delete_record,
    get_record_by_id,
    get_record_total_split_amount,
    get_records,
    update_record,
    update_record_and_splits,
)
from bagels.managers.splits import get_split_by_id, update_split
from bagels.forms.form import Form
from bagels.utils.format import format_date_to_readable
from bagels.forms.person_forms import PersonForm
from bagels.forms.record_forms import RecordForm


class DisplayMode:
    DATE = "d"
    PERSON = "p"


class Records(Static):
    BINDINGS = [
        (CONFIG.hotkeys.new, "new", "Add"),
        (CONFIG.hotkeys.delete, "delete", "Delete"),
        (CONFIG.hotkeys.edit, "edit", "Edit"),
        (CONFIG.hotkeys.home.new_transfer, "new_transfer", "Transfer"),
        (CONFIG.hotkeys.home.toggle_splits, "toggle_splits", "Toggle Splits"),
        Binding(
            CONFIG.hotkeys.home.display_by_person,
            "display_by_person",
            "Display by Person",
            show=False,
        ),
        Binding(
            CONFIG.hotkeys.home.display_by_date,
            "display_by_date",
            "Display by Date",
            show=False,
        ),
    ]

    can_focus = True
    show_splits = True
    displayMode = reactive(DisplayMode.DATE)

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="records-container", classes="module-container"
        )
        super().__setattr__("border_title", "Records")
        self.page_parent = parent
        self.record_form = RecordForm()
        self.person_form = PersonForm()

    def on_mount(self) -> None:
        self.rebuild()

    # ---------- Table builders ---------- #
    # region Table

    def rebuild(self) -> None:
        if not hasattr(self, "table"):
            return
        table = self.table
        empty_indicator: EmptyIndicator = self.query_one("#empty-indicator")
        self._initialize_table(table)
        records = get_records(**self.page_parent.filter)

        match self.displayMode:
            case DisplayMode.PERSON:
                self._build_person_view(table, records)
            case DisplayMode.DATE:
                self._build_date_view(table, records)
            case _:
                pass

        table.focus()
        if hasattr(self, "current_row_index"):
            table.move_cursor(row=self.current_row_index)
        empty_indicator.display = not table.rows

    def _initialize_table(self, table: DataTable) -> None:
        table.clear()
        table.columns.clear()
        match self.displayMode:
            case DisplayMode.PERSON:
                table.add_columns(
                    " ", "Date", "Record date", "Category", "Amount", "Paid to account"
                )
            case DisplayMode.DATE:
                table.add_columns(" ", "Category", "Amount", "Label", "Account")

    # region Date view
    def _build_date_view(self, table: DataTable, records: list) -> None:
        prev_group = None
        for record in records:
            flow_icon = self._get_flow_icon(len(record.splits) > 0, record.isIncome)

            category_string, amount_string, account_string = self._format_record_fields(
                record, flow_icon
            )
            label_string = record.label if record.label else "-"

            # Add group header based on filter type
            group_string = None
            match self.page_parent.filter["offset_type"]:
                case "year":
                    # Group by month
                    group_string = record.date.strftime("%B %Y")
                case "month":
                    # Group by week
                    week_start = record.date - timedelta(
                        days=(record.date.weekday() - CONFIG.defaults.first_day_of_week)
                        % 7
                    )
                    week_end = week_start + timedelta(days=6)

                    # Adjust week_start and week_end if they are not in the same month as record.date
                    if week_start.month != record.date.month:
                        week_start = record.date.replace(day=1)
                    if week_end.month != record.date.month:
                        last_day_of_month = (
                            record.date.replace(day=1) + timedelta(days=32)
                        ).replace(day=1) - timedelta(days=1)
                        week_end = last_day_of_month

                    group_string = f"{format_date_to_readable(week_start)} - {format_date_to_readable(week_end)}"
                case "week":
                    # Group by day
                    group_string = format_date_to_readable(record.date)
                case "day":
                    # No grouping
                    pass

            if group_string and prev_group != group_string:
                prev_group = group_string
                self._add_group_header_row(table, group_string)

            # Add main record row
            table.add_row(
                " ",
                category_string,
                amount_string,
                label_string,
                account_string,
                key=f"r-{str(record.id)}",
            )

            # Add split rows if applicable
            if record.splits and self.show_splits:
                self._add_split_rows(table, record, flow_icon)

    def _get_flow_icon(self, recordHasSplits: bool, is_income: bool) -> str:
        if recordHasSplits and not self.show_splits:
            flow_icon_positive = f"[green]=[/green]"
            flow_icon_negative = f"[red]=[/red]"
        else:
            flow_icon_positive = f"[green]{CONFIG.symbols.amount_positive}[/green]"
            flow_icon_negative = f"[red]{CONFIG.symbols.amount_negative}[/red]"
        return flow_icon_positive if is_income else flow_icon_negative

    def _format_record_fields(self, record, flow_icon: str) -> tuple[str, str]:
        if record.isTransfer:
            from_account = (
                "[italic]" + record.account.name + "[/italic]"
                if record.account.hidden
                else record.account.name
            )
            to_account = (
                "[italic]" + record.transferToAccount.name + "[/italic]"
                if record.transferToAccount.hidden
                else record.transferToAccount.name
            )
            category_string = f"{from_account} → {to_account}"
            amount_string = record.amount
            account_string = "-"
        else:
            color_tag = record.category.color.lower()
            category_string = f"[{color_tag}]{CONFIG.symbols.category_color}[/{color_tag}] {record.category.name}"

            if record.splits and not self.show_splits:
                amount_self = round(
                    record.amount - get_record_total_split_amount(record.id)
                )
                amount_string = f"{flow_icon} {amount_self}"
            else:
                amount_string = f"{flow_icon} {record.amount}"

            account_string = record.account.name

        return category_string, amount_string, account_string

    def _add_group_header_row(
        self, table: DataTable, string: str, key: str = None
    ) -> None:
        table.add_row("//", string, "", "", "", style_name="group-header", key=key)

    def _add_split_rows(self, table: DataTable, record, flow_icon: str) -> None:
        color = record.category.color.lower()
        amount_self = round(record.amount - get_record_total_split_amount(record.id), 2)
        split_flow_icon = (
            f"[red]{CONFIG.symbols.amount_negative}[/red]"
            if record.isIncome
            else f"[green]{CONFIG.symbols.amount_positive}[/green]"
        )
        line_char = f"[{color}]{CONFIG.symbols.line_char}[/{color}]"
        finish_line_char = f"[{color}]{CONFIG.symbols.finish_line_char}[/{color}]"

        for split in record.splits:
            paid_status_icon = self._get_split_status_icon(split)
            date_string = (
                Text(f"Paid {format_date_to_readable(split.paidDate)}", style="italic")
                if split.paidDate
                else Text("-")
            )

            table.add_row(
                " ",
                f"{line_char} {paid_status_icon} {split.person.name}",
                f"{split_flow_icon} {split.amount}",
                date_string,
                split.account.name if split.account else "-",
                key=f"s-{str(split.id)}",
            )

        # Add net amount row
        table.add_row(
            "",
            f"{finish_line_char} Self total",
            f"= {amount_self}",
            "",
            "",
            style_name="net",
        )

    def _get_split_status_icon(self, split) -> str:
        if split.isPaid:
            return f"[green]{CONFIG.symbols.split_paid}[/green]"
        else:
            return f"[grey]{CONFIG.symbols.split_unpaid}[/grey]"

    # region Person view
    def _build_person_view(self, table: DataTable, _) -> None:
        persons = get_persons_with_splits(**self.page_parent.filter)

        # Display each person and their splits
        for person in persons:
            if person.splits:  # Person has splits for this month
                # Add person header
                self._add_group_header_row(
                    table, person.name, key=f"p-{str(person.id)}"
                )

                # Add splits for this person
                for split in person.splits:
                    record = split.record
                    paid_icon = (
                        f"[green]{CONFIG.symbols.split_paid}[/green]"
                        if split.isPaid
                        else f"[red]{CONFIG.symbols.split_unpaid}[/red]"
                    )
                    date = (
                        format_date_to_readable(split.paidDate)
                        if split.paidDate
                        else "Not paid"
                    )
                    record_date = format_date_to_readable(record.date)
                    category = f"[{record.category.color.lower()}]{CONFIG.symbols.category_color}[/{record.category.color.lower()}] {record.category.name}"
                    amount = (
                        f"[red]{CONFIG.symbols.amount_negative}[/red] {split.amount}"
                        if record.isIncome
                        else f"[green]{CONFIG.symbols.amount_positive}[/green] {split.amount}"
                    )
                    account = f"→ {split.account.name}" if split.account else "-"

                    table.add_row(
                        " ",
                        f"{paid_icon} {date}",
                        record_date,
                        category,
                        amount,
                        account,
                        key=f"s-{split.id}",
                    )

    # region Helpers
    # -------------- Helpers ------------- #

    # region Callbacks
    # ------------- Callbacks ------------ #

    def on_data_table_row_highlighted(self, event: DataTable.RowHighlighted) -> None:
        current_row_index = event.cursor_row
        if event.row_key and event.row_key.value:
            self.current_row = event.row_key.value
            self.current_row_index = current_row_index
        else:
            self.current_row = None
            self.current_row_index = None

    def watch_displayMode(self, displayMode: DisplayMode) -> None:
        self.query_one("#display-date").classes = (
            "selected" if displayMode == DisplayMode.DATE else ""
        )
        self.query_one("#display-person").classes = (
            "selected" if displayMode == DisplayMode.PERSON else ""
        )

    def action_toggle_splits(self) -> None:
        self.show_splits = not self.show_splits
        self.rebuild()

    def action_display_by_person(self) -> None:
        self.displayMode = DisplayMode.PERSON
        self.rebuild()

    def action_display_by_date(self) -> None:
        self.displayMode = DisplayMode.DATE
        self.rebuild()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        match event.button.id:
            case "prev-month":
                self.page_parent.action_prev_month()
            case "next-month":
                self.page_parent.action_next_month()
            case "display-date":
                self.action_display_by_date()
            case "display-person":
                self.action_display_by_person()
            case _:
                pass

    # region cud

    def action_new(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_record_and_splits(result["record"], result["splits"])
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Record created",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()

        date = format_date_to_readable(self.page_parent.mode["date"])
        account_name = self.page_parent.mode["accountId"]["default_value_text"]
        type = "Income" if self.page_parent.mode["isIncome"] else "Expense"
        self.app.push_screen(
            RecordModal(
                f"New {type} on {account_name} for {date}",
                form=self.record_form.get_form(self.page_parent.mode),
                splitForm=Form(),
            ),
            callback=check_result,
        )

    def action_edit(self) -> None:
        if not (hasattr(self, "current_row") and self.current_row):
            self.app.notify(
                title="Error", message="Nothing selected", severity="error", timeout=2
            )
            self.app.bell()
            return
        # ----------------- - ---------------- #
        type = self.current_row.split("-")[0]
        id = self.current_row.split("-")[1]

        # ----------------- - ---------------- #
        def check_result_records(result) -> None:
            if result:
                try:
                    if result.get("record"):  # if not editing a transfer:
                        update_record_and_splits(id, result["record"], result["splits"])
                    else:
                        update_record(id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Record updated",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()
            else:
                self.app.notify(
                    title="Discarded",
                    message=f"Record not updated",
                    severity="warning",
                    timeout=3,
                )

        def check_result_person(result) -> None:
            if result:
                try:
                    update_person(id, result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Person updated",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()
            else:
                self.app.notify(
                    title="Discarded",
                    message=f"Person not updated",
                    severity="warning",
                    timeout=3,
                )

        # ----------------- - ---------------- #
        match type:
            case "r":
                record = get_record_by_id(id)
                if not record:
                    self.app.notify(
                        title="Error",
                        message="Record not found",
                        severity="error",
                        timeout=2,
                    )
                    return
                if record.isTransfer:
                    self.app.push_screen(
                        TransferModal(record), callback=check_result_records
                    )
                else:
                    filled_form, filled_splits = self.record_form.get_filled_form(
                        record.id
                    )
                    self.app.push_screen(
                        RecordModal(
                            "Edit Record",
                            form=filled_form,
                            splitForm=filled_splits,
                            isEditing=True,
                        ),
                        callback=check_result_records,
                    )
            case "s":
                split = get_split_by_id(id)
                if split.isPaid:
                    split_data = {"accountId": None, "isPaid": False, "paidDate": None}
                    update_split(id, split_data)
                    self.app.notify(
                        title="Reverted split",
                        message=f"Marked this split as unpaid",
                        severity="information",
                        timeout=3,
                    )
                else:
                    split_data = {
                        "accountId": self.page_parent.mode["accountId"][
                            "default_value"
                        ],
                        "isPaid": True,
                        "paidDate": datetime.now(),
                    }
                    update_split(id, split_data)
                    self.app.notify(
                        title="Completed split",
                        message=f"With account {self.page_parent.mode['accountId']['default_value_text']} today",
                        severity="information",
                        timeout=3,
                    )
                self.page_parent.rebuild()
            case "p":
                person = get_person_by_id(id)
                if not person:
                    self.app.notify(
                        title="Error",
                        message="Person not found",
                        severity="error",
                        timeout=2,
                    )
                    return
                self.app.push_screen(
                    InputModal(
                        "Edit Person", form=self.person_form.get_filled_form(person.id)
                    ),
                    callback=check_result_person,
                )
            case _:
                pass

    def action_delete(self) -> None:
        if not (hasattr(self, "current_row") and self.current_row):
            self.app.notify(
                title="Error", message="Nothing selected", severity="error", timeout=2
            )
            self.app.bell()
            return
        # ----------------- - ---------------- #
        type = self.current_row.split("-")[0]
        id = self.current_row.split("-")[1]

        if type == "s":
            self.app.notify(
                title="Error",
                message="You cannot delete or add splits to a record after creation.",
                severity="error",
                timeout=2,
            )
            return

        # ----------------- - ---------------- #
        def check_delete(result) -> None:
            if result:
                delete_record(id)
                self.app.notify(
                    title="Success",
                    message=f"Record deleted",
                    severity="information",
                    timeout=3,
                )
                self.page_parent.rebuild()

        # ----------------- - ---------------- #
        match type:
            case "r":
                self.app.push_screen(
                    ConfirmationModal("Are you sure you want to delete this record?"),
                    callback=check_delete,
                )
            case "s":
                self.app.push_screen(
                    ConfirmationModal("Are you sure you want to delete this split?"),
                    callback=check_delete,
                )
            case _:
                pass

    def action_new_transfer(self) -> None:
        def check_result(result) -> None:
            if result:
                try:
                    create_record(result)
                except Exception as e:
                    self.app.notify(
                        title="Error", message=f"{e}", severity="error", timeout=10
                    )
                else:
                    self.app.notify(
                        title="Success",
                        message=f"Record created",
                        severity="information",
                        timeout=3,
                    )
                    self.page_parent.rebuild()
            else:
                self.app.notify(
                    title="Discarded",
                    message=f"Record not updated",
                    severity="warning",
                    timeout=3,
                )

        self.app.push_screen(TransferModal(), callback=check_result)

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        with Container(classes="selectors"):
            displayContainer = Container(classes="display-selector")
            displayContainer.border_title = "Display by:"
            displayContainer.border_subtitle = f"{CONFIG.hotkeys.home.display_by_date} {CONFIG.hotkeys.home.display_by_person}"
            with displayContainer:
                yield Button(f"Date", id="display-date")
                yield Button(f"Person", id="display-person")
        self.table = DataTable(
            id="records-table",
            cursor_type="row",
            cursor_foreground_priority=True,
            zebra_stripes=True,
            additional_classes=["datatable--net-row", "datatable--group-header-row"],
        )
        yield self.table
        yield EmptyIndicator("No entries")



import getpass
import socket

from rich.text import Text


def get_user_host_string() -> Text:
    try:
        username = getpass.getuser()
        hostname = socket.gethostname()
        return f"{username}@{hostname}"
    except Exception:
        return "unknown@unknown"
